"""
zmail.server
~~~~~~~~~~~~
This module provides a MailServer object to communicate with mail server.
"""

import smtplib
import poplib
import logging

from zmail.exceptions import InvalidProtocol
from zmail.info import get_supported_server_info
from zmail.structures import CaseInsensitiveDict
from zmail.message import mail_encode, mail_decode
from zmail.settings import __local__
from zmail.message import parse_header_shortcut

# Fix poplib bug.
poplib._MAXLINE = 4096

logger = logging.getLogger('zmail')


class MailServer:
    """High-level MailServer API includes SMTPServer and POPServer."""

    def __init__(self, user: str, password: str, smtp_host: str, smtp_port: int, pop_host: str, pop_port: int,
                 smtp_ssl: bool, pop_ssl: bool, smtp_tls: bool, pop_tls: bool, timeout=60,
                 auto_add_from=True, auto_add_to=True):
        self.user = user
        self.password = password
        self.timeout = timeout

        self.smtp_host = smtp_host
        self.smtp_port = smtp_port
        self.smtp_ssl = smtp_ssl
        self.smtp_tls = smtp_tls

        self.pop_host = pop_host
        self.pop_port = pop_port
        self.pop_ssl = pop_ssl
        self.pop_tls = pop_tls

        self.auto_add_from = auto_add_from
        self.auto_add_to = auto_add_to

    def send_mail(self, recipients, message, timeout=60):
        """"Send email."""
        message = CaseInsensitiveDict(message)

        if self.auto_add_from and message.get('from') is None:
            message['from'] = '{}<{}>'.format(self.user.split("@")[0], self.user)

        recipients = recipients if isinstance(recipients, (list, tuple)) else (recipients,)

        host, port, ssl, tls = get_supported_server_info(self.user, 'smtp')
        host = self.smtp_host if self.smtp_host is not None else host
        port = self.smtp_port if self.smtp_port is not None else port
        ssl = self.smtp_ssl if self.smtp_ssl is not None else ssl
        tls = self.pop_tls if self.pop_tls is not None else tls

        logger.info('Prepare login into {}:{} ssl:{} tls:{}.'.format(host, port, ssl, tls))

        server = SMTPServer(host, port, self.user, self.password)

        message = mail_encode(message)

        if ssl:
            server.send_ssl(recipients, message, timeout, self.auto_add_to)
        else:
            server.send(recipients, message, timeout, self.auto_add_to, tls)

        return True

    def stat(self) -> tuple:
        """Get mailbox status."""
        server = self._init_pop3()
        status = server.stat()
        server.logout()

        return status

    def get_mail(self, which: int):
        """Get a mail from mailbox."""
        server = self._init_pop3()

        mail = server.get_mail(which)

        server.logout()

        return mail_decode(mail, which)

    def get_mails(self, subject=None, after=None, before=None, sender=None):
        """Get a list of mails from mailbox."""
        info = self.get_info()
        mail_id = []

        for index, mail in enumerate(info):
            if self._match(parse_header_shortcut(mail), subject, after, before, sender):
                mail_id.append(index + 1)

        server = self._init_pop3()

        mail_id.sort()
        mail_as_bytes_list = server.get_mails(mail_id)
        server.logout()

        return [mail_decode(mail_as_bytes, mail_id[index]) for index, mail_as_bytes in enumerate(mail_as_bytes_list)]

    def get_latest(self):
        """Get latest mail in mailbox."""
        server = self._init_pop3()

        latest_num = server.stat()[0]
        mail = server.get_mail(latest_num)

        server.logout()

        return mail_decode(mail, latest_num)

    def get_info(self) -> list:
        """Get all mails information.include(subject,from,to,date)"""
        server = self._init_pop3()

        result = server.get_headers()

        server.logout()

        return result

    def smtp_able(self):
        host, port, ssl, tls = get_supported_server_info(self.user, 'smtp')
        host = self.smtp_host if self.smtp_host is not None else host
        port = self.smtp_port if self.smtp_port is not None else port
        ssl = self.smtp_ssl if self.smtp_ssl is not None else ssl
        tls = self.pop_tls if self.pop_tls is not None else tls

        logger.info('Prepare login into {}:{} ssl:{} tls:{}.'.format(host, port, ssl, tls))

        server = SMTPServer(host, port, self.user, self.password)
        return server.login_able(ssl, tls, self.timeout)

    def pop_able(self):
        try:
            server = self._init_pop3()
            server.logout()
            return True
        except Exception as e:
            logger.warning('Login pop3 error :{}'.format(e))
            return False

    def _init_pop3(self):
        """Initiate POP3 server."""
        host, port, ssl, tls = get_supported_server_info(self.user, 'pop3')
        host = self.pop_host if self.pop_host is not None else host
        port = self.pop_port if self.pop_port is not None else port
        ssl = self.pop_ssl if self.pop_ssl is not None else ssl
        tls = self.pop_tls if self.pop_tls is not None else tls

        logger.info('Prepare login into {}:{} ssl:{}.'.format(host, port, ssl))
        server = POP3Server(host, port, self.user, self.password, ssl, tls, self.timeout)

        return server

    @staticmethod
    def _match(mail, subject=None, after=None, before=None, sender=None):
        """Match all conditions."""
        _subject = mail.get('subject', 'None')
        _sender = mail.get('from', 'None')

        _date = mail['date'].split(' ')[0]
        _time = tuple(map(lambda x: int(x), _date.split('-')))

        if subject and _subject.find(subject) == -1:
            return False

        if sender and _sender.find(sender) == -1:
            return False

        if before:
            # bf_year, bf_month, bf_day.
            before_time = tuple(map(lambda x: int(x), before.split('-')))
            for i in range(0, 3):
                if before_time[i] < _time[i]:
                    return False
                elif before_time[i] > _time[i]:
                    break
        if after:
            after_time = tuple(map(lambda x: int(x), after.split('-')))
            for i in range(0, 3):
                if after_time[i] > _time[i]:
                    return False
                elif after_time[i] < _time[i]:
                    break
        return True


class SMTPServer:
    """Base SMTPServer, which encapsulates python3 standard library to a SMTPServer."""

    def __init__(self, host, port, user, password):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def send_ssl(self, recipients, message, timeout, auto_add_to):
        with smtplib.SMTP_SSL(self.host, self.port, __local__, timeout=timeout) as server:
            server.login(self.user, self.password)
            for recipient in recipients:
                if auto_add_to and message.get('to') is None:
                    message['To'] = '{}<{}>'.format(recipient.split("@")[0], recipient)
                server.sendmail(self.user, recipient, message.as_string())

    def send(self, recipients, message, timeout, auto_add_to, tls=True):
        with smtplib.SMTP(self.host, self.port, __local__, timeout=timeout) as server:
            if tls:
                server.ehlo()
                server.starttls()
                server.ehlo()
            server.login(self.user, self.password)
            for recipient in recipients:
                if auto_add_to and message.get('to') is None:
                    message['To'] = '{}<{}>'.format(recipient.split("@")[0], recipient)
                server.sendmail(self.user, recipient, message.as_string())

    def login_able(self, use_ssl, use_tls, timeout):
        try:
            if use_ssl:
                with smtplib.SMTP_SSL(self.host, self.port, __local__, timeout=timeout) as server:
                    server.login(self.user, self.password)
            else:
                with smtplib.SMTP(self.host, self.port, __local__, timeout=timeout) as server:
                    if use_tls:
                        server.ehlo()
                        server.starttls()
                        server.ehlo()
                    server.login(self.user, self.password)
        except Exception as e:
            logger.warning('Login SMTP error :{}'.format(e))
            return False
        return True


class POP3Server:
    """Base POP3Server, which encapsulates python3 standard library to a SMTPServer."""

    def __init__(self, host: str, port: int, user: str, password: str, ssl=False, tls=False, timeout=60):
        self.user = user
        self.password = password

        self.pop3 = poplib.POP3_SSL(host, port, timeout=timeout) if ssl else poplib.POP3(host, port, timeout=timeout)

        if tls and not ssl:
            self.pop3.stls()
        elif tls and ssl:
            raise InvalidProtocol('Can not used ssl and tls together.')

        self.login()

    def login(self):
        """Note: the mailbox on the server is locked until logout() is called."""
        self.pop3.user(self.user)
        self.pop3.pass_(self.password)

    def logout(self):
        """Quit pop3 server."""
        self.pop3.quit()

    def stat(self) -> tuple:
        """Get mailbox status. The result is a tuple of 2 integers: (message count, mailbox size)."""
        return self.pop3.stat()

    def get_header(self, which) -> list:
        """Use 'top' to get mail headers."""
        return self.pop3.top(which, 0)[1]

    def get_headers(self) -> list:
        """Get all mails headers."""
        num = self.stat()[0]
        result = []
        for count in range(1, num + 1):
            header_as_bytes = self.get_header(count)
            result.append(header_as_bytes)
        return result

    def get_mail(self, which) -> list:
        """Get a mail by its id."""
        return self.pop3.retr(which)[1]

    def get_mails(self, which_list: list) -> list:
        """Get a list of mails by its id."""
        return [self.pop3.retr(which)[1] for which in which_list]



"""
zmail.info
~~~~~~~~~~
This module provide supported server information.


'Server_provider_address':{
    'protocol':('protocol_server_address', port, use_ssl,use_tls),
}
"""
SUPPORTED_SERVER = {
    '163.com': {
        'smtp': ('smtp.163.com', 994, True, False),
        'pop3': ('pop.163.com', 995, True, False),
        'imap': ('imap.163.com', 993, True, False)

    },
    '126.com': {
        'smtp': ('smtp.126.com', 994, True, False),
        'pop3': ('pop.126.com', 995, True, False),
        'imap': ('imap.126.com', 993, True, False)

    },
    'yeah.net': {
        'smtp': ('smtp.yeah.net', 994, True, False),
        'pop3': ('pop.yeah.net', 995, True, False),
        'imap': ('imap.yeah.net', 993, True, False)

    },
    'qq.com': {
        'smtp': ('smtp.qq.com', 465, True, False),
        'pop3': ('pop.qq.com', 995, True, False),
    },
    'gmail.com': {
        'smtp': ('smtp.gmail.com', 587, False, True),
        'pop3': ('pop.gmail.com', 995, True, False),
    },
    'sina.com': {
        'smtp': ('smtp.sina.com', 465, True, False),
        'pop3': ('pop.sina.com', 995, True, False),
    },
    'outlook.com': {
        'smtp': ('smtp-mail.outlook.com', 587, False, True),
        'pop3': ('pop.outlook.com', 995, True, False),
    },

}

SUPPORTED_ENTERPRISE_SERVER_CONFIG = {
    'qq': {
        'smtp_host': 'smtp.exmail.qq.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.exmail.qq.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False

    },
    'aliyun': {
        'smtp_host': 'smtp.mxhichina.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop3.mxhichina.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False
    },

}


def get_supported_server_info(mail_address: str, protocol: str) -> tuple:
    """Use user address to get server address and port.

    :param mail_address: str
    :param protocol: str
    :return: ('protocol_server_address', port, use_ssl)
    """
    provider = mail_address.split('@')[1]

    if provider in SUPPORTED_SERVER:
        server_info = SUPPORTED_SERVER[provider]
        if protocol in server_info:
            return server_info[protocol]

    if protocol == 'smtp':
        return 'smtp.' + provider, 465, True, False
    elif protocol == 'pop3':
        return 'pop3.' + provider, 995, True, False


def get_enterprise_server_config(config: str):
    """Get user-defined config.
    :param config: str
    :return: ('protocol_server_address', port, use_ssl)
    """
    if config in SUPPORTED_ENTERPRISE_SERVER_CONFIG:
        return SUPPORTED_ENTERPRISE_SERVER_CONFIG[config]
    return False



"""
zmail.message
~~~~~~~~~~~~
This module provides functions to handles MIME object.
"""
import os
import re
import mimetypes
import logging
import quopri
import base64

from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.encoders import encode_base64
from email.header import decode_header

from .utils import get_abs_path, make_iterable
from .structures import CaseInsensitiveDict

logger = logging.getLogger('zmail')


def mail_encode(message: CaseInsensitiveDict):
    """Convert a dict to a MIME obj."""
    logger.info('Encoding mail to MIME obj.')

    # Create MIMEMultipart.
    msg = MIMEMultipart()

    # Set basic email elements.
    for k, v in message.lower_items():
        if k in ('from', 'to', 'subject') and v:
            msg[k.capitalize()] = v

    # Set extra parameters.
    for k in message:
        if k.lower() not in ('from', 'to', 'subject', 'attachments', 'content'):
            msg[k] = message[k]

    # Set mail content.
    if 'content_html' in message:
        msg.attach(MIMEText('%s' % message['content_html'], 'html', 'utf-8'))
    if 'content' in message:
        _message = make_iterable(message['content'])
        _combine_message = ''
        for i in _message:
            _combine_message += i
        msg.attach(MIMEText('{}'.format(_combine_message), 'plain', 'utf-8'))

    # Set attachments.
    if 'attachments' in message and message['attachments']:
        attachments = make_iterable(message['attachments'])
        for attachment in attachments:
            logger.info('Loading %s', attachment)
            attachment = get_abs_path(attachment)
            part = _get_attachment_part(attachment)
            msg.attach(part)

    return msg


def mail_decode(mail_as_bytes, which=1):
    """Convert a MIME bytes to dict."""
    return MailDecode(mail_as_bytes, which).result


def parse_header(mail_as_bytes, *args):
    """Parse mail header then return a dictionary include basic elements in email.

    :param mail_as_bytes: mail as bytes
    :param args: extra parameters
    :return: dict {'header_as_str':'value_as_bytes'}
    """
    email_headers = {}

    # Parse headers.
    extra_header = {}
    for i in args:
        extra_header[i] = None
    extra_pattern = rb"""[\s]?=["'\s]?([^"';\\\r\n\s]+)["';\r\n\s]?"""

    part = b''
    header = b''

    for line in mail_as_bytes:
        # Get extra_header if possible.
        for k, v in extra_header.items():
            if v is None and re.search(bytes(k, encoding='utf8') + bytes(extra_pattern), line):
                extra_header[k] = re.search(bytes(k, encoding='utf8') + bytes(extra_pattern), line).group(1)

        # Parsing header.
        if re.fullmatch(rb'([0-9a-zA-Z-]+):\s?([^\r\n]+)(\r\n)?', line):
            # Line is a header.
            if part and header:
                email_headers[header.decode()] = part
            header, part = re.search(rb'([0-9a-zA-Z-]+):\s?([^\r\n]+)(\r\n)?', line).group(1, 2)
        elif re.fullmatch(rb'[\t]?([^\r\n]+)(\r\n)?', line):
            # Line is a part of header.
            _part = re.search(rb'[\t]?([^\r\n]+)(\r\n)?', line).group(1)
            part += _part
        elif line in (b'\r\n', b''):
            # Mail header boundary.
            if part and header:
                email_headers[header.decode()] = part
            break
        else:
            raise Exception('Parsing error:%s' % line)

    # Add extra header if possible.
    email_headers = {**email_headers, **extra_header}

    return email_headers


def _get_attachment_part(file):
    """According to file-type return a prepared attachment part."""
    name = os.path.split(file)[1]
    file_type = mimetypes.guess_type(name)[0]

    if file_type is None:
        logger.warning('Could not guess %s type, use application type instead.', file)
        file_type = 'application/octet-stream'

    main_type, sub_type = file_type.split('/')

    if main_type == 'text':
        with open(file, 'r') as f:
            part = MIMEText(f.read())
            part['Content-Disposition'] = 'attachment;filename="%s"' % name

    elif main_type in ('image', 'audio'):
        with open(file, 'rb') as f:
            part = MIMEImage(f.read(), _subtype=sub_type) if main_type == 'image' else \
                MIMEAudio(f.read(), _subtype=sub_type)
            part['Content-Disposition'] = 'attachment;filename="%s"' % name
    else:
        with open(file, 'rb') as f:
            part = MIMEBase(main_type, sub_type)
            part.set_payload(f.read())
            part['Content-Disposition'] = 'attachment;filename="%s"' % name
            encode_base64(part)

    return part


def _fmt_date(date_as_string):
    """Format mail header Date for humans."""
    _month = {
        'Jan': 1,
        'Feb': 2,
        'Mar': 3,
        'Apr': 4,
        'May': 5,
        'Jun': 6,
        'Jul': 7,
        'Aug': 8,
        'Sep': 9,
        'Oct': 10,
        'Nov': 11,
        'Dec': 12,
    }
    pattern_one = r'(\w+),\s+([0-9]+)\s+(\w+)\s+([0-9]+)\s+([:0-9]+)\s+(.+)'
    pattern_two = r'([0-9])\s+([\w]+)\s+([0-9]+)\s+([:0-9]+)\s+(.+)'
    match_one = re.search(pattern_one, date_as_string)
    match_two = re.search(pattern_two, date_as_string)
    if match_one:
        week, day, month_as_string, year, now_time, time_zone = match_one.groups()
        month = [v for k, v in _month.items() if month_as_string.lower() == k.lower()][0]
        return '{}-{}-{} {} {}'.format(int(year), int(month), int(day), now_time, time_zone)
    elif match_two:
        day, month_as_string, year, now_time, time_zone = match_two.groups()
        month = [v for k, v in _month.items() if month_as_string.lower() == k.lower()][0]
        return '{}-{}-{} {} {}'.format(int(year), int(month), int(day), now_time, time_zone)
    else:
        logger.warning('Can not parse Date:{}'.format(date_as_string))
        return date_as_string


def divide_into_parts(bytes_list, sep):
    """Divide a list of bytes to parts.
    """
    is_part = False
    result = []

    part = []
    for i in bytes_list:
        if i.find(sep) > -1:
            is_part = True
            if part:
                result.append(part)
            part = []
            continue

        if is_part:
            part.append(i)

    return result


def parse_header_shortcut(mail_as_bytes):
    """Shortcut for parse mail headers(only)."""

    def _decode_header(header_as_string):
        result = ''

        for i in decode_header(header_as_string):
            _bytes, charset = i
            charset = charset or 'utf-8'
            if isinstance(_bytes, bytes):
                try:
                    result += _bytes.decode(charset)
                except UnicodeDecodeError:
                    logger.warning('UnicodeDecodeError:{} {}'.format(_bytes, charset))
            elif isinstance(_bytes, str):
                result += _bytes
            else:
                raise Exception('Can not decode header:{}'.format(header_as_string))
        return result

    # Set default values.
    headers = CaseInsensitiveDict()
    for i in ('subject', 'from', 'to', 'date', 'content-type', 'boundary'):
        headers.setdefault(i, None)
    headers['charset'] = 'utf-8'
    # Update values by parse headers.
    _parsed_headers = parse_header(mail_as_bytes, 'boundary', 'charset', 'content-transfer-encoding')
    parsed_headers = _parsed_headers.copy()
    for k, v in _parsed_headers.items():
        if v is None:
            parsed_headers.pop(k)
    headers.update(parsed_headers)
    # Convert charset to str for subsequent parsing.
    if isinstance(headers['charset'], bytes):
        headers['charset'] = headers['charset'].decode()

    # Decode values use its charset.
    # 'boundary' need to be bytes for subsequent parsing.
    # 'charset' is already str.
    for k, v in headers.lower_items():
        if headers[k] and k not in ('boundary', 'charset'):
            try:
                headers[k] = v.decode(headers['charset'])
            except UnicodeDecodeError:
                if k in ('subject', 'from', 'to'):
                    try:
                        headers[k] = v.decode('utf-8')
                    except UnicodeDecodeError:
                        logger.critical(
                            'Can not decode basic "{}",Use {} instead, raw:{}'.format(k, 'Unknown%s' % k,
                                                                                      headers[k]))
                        headers[k] = 'Unknown%s' % k
                else:
                    logger.warning('Can not decode {} {}'.format(k, headers['charset']))
    # Decode headers for read.
    # Usually, subject|from|to need to be decode again.
    for k in ('subject', 'from', 'to'):
        if headers[k]:
            headers[k] = _decode_header(headers[k])

    # Decode 'date' for read.
    if headers.get('date'):
        headers['date'] = _fmt_date(headers['date'])

    # Warning failed to catch basic mail elements .
    for i in ('subject', 'date', 'from', 'to', 'content-type'):
        if headers[i] is None:
            logger.warning("Can not get element '%s' in this mail." % i)
    return headers


class MailDecode:
    def __init__(self, mail_as_bytes, which):
        """
        Mail content:
        - headers content (generated from mail headers)
        'subject'                   basic. default None.
        'from'                      basic. default None.
        'to'                        basic. default None.
        'date'                      basic. default None.
        'content-type'              basic. default None.
        'content-transfer-encoding' extra. default None.
        'boundary'                  extra. default None.
        'charset'                   extra. default 'utf-8'.

        - result content (generated from mail parsing)
        'id'                        extra. from mail server.
        'content'                   extra. default None.
        'content-html'              extra. default None.
        'attachments'               extra. default None.
        'headers'                   basic. all headers can be parsed from mail, include processed headers content.

        - raw (mail itself)
        'raw'                       basic. mail itself as raw.
        """
        self.mail_as_bytes = mail_as_bytes

        # Set default values.
        headers = CaseInsensitiveDict()
        for i in ('subject', 'from', 'to', 'date', 'content-type', 'boundary'):
            headers.setdefault(i, None)
        headers['charset'] = 'utf-8'
        # Update values by parse headers.
        _parsed_headers = parse_header(mail_as_bytes, 'boundary', 'charset', 'content-transfer-encoding')
        parsed_headers = _parsed_headers.copy()
        for k, v in _parsed_headers.items():
            if v is None:
                parsed_headers.pop(k)
        headers.update(parsed_headers)
        # Convert charset to str for subsequent parsing.
        if isinstance(headers['charset'], bytes):
            headers['charset'] = headers['charset'].decode()

        # Decode values use its charset.
        # 'boundary' need to be bytes for subsequent parsing.
        # 'charset' is already str.
        for k, v in headers.lower_items():
            if headers[k] and k not in ('boundary', 'charset'):
                try:
                    headers[k] = v.decode(headers['charset'])
                except UnicodeDecodeError:
                    if k in ('subject', 'from', 'to'):
                        try:
                            headers[k] = v.decode('utf-8')
                        except UnicodeDecodeError:
                            logger.critical(
                                'Can not decode basic "{}",Use {} instead, raw:{}'.format(k, 'Unknown%s' % k,
                                                                                          headers[k]))
                            headers[k] = 'Unknown%s' % k
                    else:
                        logger.warning('Can not decode {} {}'.format(k, headers['charset']))
        # Decode headers for read.
        # Usually, subject|from|to need to be decode again.
        for k in ('subject', 'from', 'to'):
            if headers[k]:
                headers[k] = self._decode_header(headers[k])

        # Decode 'date' for read.
        if headers.get('date'):
            headers['date'] = _fmt_date(headers['date'])

        # Warning failed to catch basic mail elements .
        for i in ('subject', 'date', 'from', 'to', 'content-type'):
            if headers[i] is None:
                logger.warning("Can not get element '%s' in this mail." % i)

        # Headers init is complete.
        self.headers = headers

        # Make result init.
        self.result = dict()
        for k in ('subject', 'from', 'to', 'date', 'content-type', 'boundary', 'charset'):
            self.result[k] = headers.get(k)
        for k in ('attachments', 'content', 'content_html'):
            self.result.setdefault(k, None)

        self.result['id'] = which
        self.result['headers'] = headers
        self.result['raw'] = mail_as_bytes

        # Get body_as_bytes
        self.body_as_bytes = None
        for line in mail_as_bytes:
            if line in (b'\r\n', b''):
                self.body_as_bytes = mail_as_bytes[mail_as_bytes.index(line) + 1:]
        assert self.body_as_bytes is not None, 'Can not divide headers and body.'

        # Set mail attributes.
        self.content_transfer_encoding = headers.get('content-transfer-encoding')
        try:
            self.content_type = re.search(r'\s?([a-z0-9]+/[a-z0-9]+)\s?;.+', headers['content-type']).group(1)
        except AttributeError:
            self.content_type = ''
        self.charset = headers.get('charset')
        self.boundary = headers.get('boundary')

        # Attribute check.
        if self.content_transfer_encoding is not None and self.content_transfer_encoding.lower() not in (
                'quoted-printable', '8bit', '7bit', 'base64'):
            raise Exception("Unknown encoding:{}".format(self.content_transfer_encoding))

        if self.is_multiple_part():
            self.multiple_part_decode()
        else:
            _content, _content_html = self.one_part_decode()
            self.result['content'], self.result['content_html'] = _content, _content_html

    @staticmethod
    def _decode_header(header_as_string):
        result = ''

        for i in decode_header(header_as_string):
            _bytes, charset = i
            charset = charset or 'utf-8'
            if isinstance(_bytes, bytes):
                try:
                    result += _bytes.decode(charset)
                except UnicodeDecodeError:
                    logger.warning('UnicodeDecodeError:{} {}'.format(_bytes, charset))
            elif isinstance(_bytes, str):
                result += _bytes
            else:
                raise Exception('Can not decode header:{}'.format(header_as_string))
        return result

    def is_multiple_part(self):
        return self.headers.get('boundary')

    def multiple_part_decode(self):
        attachments = []
        content = []
        content_html = []

        parts = divide_into_parts(self.body_as_bytes, self.boundary)

        for part in parts:
            # Init headers.
            part_headers = CaseInsensitiveDict()
            part_headers.update(parse_header(part, 'charset'))
            part_headers['charset'] = part_headers['charset'] if part_headers.get('charset') else 'utf-8'
            if isinstance(part_headers['charset'], bytes):
                part_headers['charset'] = part_headers['charset'].decode()
            # Convert headers to string by its charset.
            for k, v in part_headers.lower_items():
                if part_headers[k] and k != 'charset':
                    try:
                        part_headers[k] = v.decode(part_headers['charset'])
                    except UnicodeDecodeError:
                        logger.warning('UnicodeDecodeError {} {} {}'.format(k, v, part_headers['charset']))
            # Divide body and headers
            part_body_as_bytes = None
            for line in part:
                if line in (b'\r\n', b''):
                    part_body_as_bytes = part[part.index(line) + 1:]
            # Empty part.
            if part_body_as_bytes is None:
                logger.info('Detect an empty part.')
                continue
            # Set part headers.
            content_transfer_encoding = part_headers.get('content-transfer-encoding')
            content_disposition = part_headers.get('content-disposition')
            try:
                content_type = re.search(r'\s?([a-z0-9]+/[a-z0-9]+)\s?;.+', part_headers['content-type']).group(1)
            except AttributeError:
                content_type = ''
            charset = part_headers.get('charset')

            # Only 3 situations: attachments,content,content_html
            # Is attachment.
            if content_disposition and content_disposition.find('attachment') > -1:
                attachment = []
                filename = re.search(r"""attachment;[\s]?filename=[\s"']?([^"';]+)""", content_disposition).group(1)
                attachment.append(filename + ';%s' % content_type)

                # Add body.
                is_body = False
                for line in part:
                    if line == b'':
                        is_body = True
                        continue
                    if is_body:
                        if content_transfer_encoding.find('base64') > -1:
                            attachment.append(base64.b64decode(line))
                        else:
                            attachment.append(line)
                attachments.append(attachment)
            elif content_type in ('text/plain', 'text/html'):
                _content, _content_html = self.part_decode(part_body_as_bytes, content_transfer_encoding, content_type,
                                                           charset)
                for i in _content:
                    content.append(i)
                for i in _content_html:
                    content_html.append(i)
            else:
                logger.warning('Can not parse:{}'.format(content_type))

        self.result['content'] = content
        self.result['content_html'] = content_html
        self.result['attachments'] = attachments

    def one_part_decode(self):

        _content = []
        if self.content_transfer_encoding is None or self.content_transfer_encoding.lower() not in (
                'base64', 'quoted-printable'):
            for i in self.body_as_bytes:
                _content.append(i.decode(self.charset))
        if self.content_transfer_encoding == 'base64':
            _temp = b''
            for i in self.body_as_bytes:
                _temp += i
            _temp = base64.b64decode(_temp).decode(self.charset)
            _content.append(_temp)
        elif self.content_transfer_encoding == 'quoted-printable':
            _temp = b''
            for i in self.body_as_bytes:
                _temp += (quopri.decodestring(i))
            _temp = _temp.decode(self.charset)
            _content.append(_temp)
        if self.content_type == 'text/plain':
            return _content, None
        elif self.content_type == 'text/html':
            return None, _content
        else:
            raise Exception('Unknown content-type:{}'.format(self.content_type))

    @staticmethod
    def part_decode(body_as_bytes, content_transfer_encoding, content_type, charset):
        """Decode 'text/html' & 'text/plain' . """
        _content = []
        if content_transfer_encoding is None or content_transfer_encoding.lower() not in (
                'base64', 'quoted-printable'):
            for i in body_as_bytes:
                _content.append(i.decode(charset))
        if content_transfer_encoding == 'base64':
            _temp = b''
            for i in body_as_bytes:
                _temp += i
            _temp = base64.b64decode(_temp).decode(charset)
            _content.append(_temp)
        elif content_transfer_encoding == 'quoted-printable':
            _temp = b''
            for i in body_as_bytes:
                _temp += (quopri.decodestring(i))
            _temp = _temp.decode(charset)
            _content.append(_temp)
        if content_type == 'text/plain':
            return _content, []
        elif content_type == 'text/html':
            return [], _content
        else:
            raise Exception('Unknown content-type:{}'.format(content_type))



"""
zmail.api
~~~~~~~~~~~~
This module implements the zmail API.
"""
import logging

from .server import MailServer
from .message import mail_decode
from .utils import get_attachment, get_html, show, read, save
from .info import get_enterprise_server_config

logger = logging.getLogger('zmail')

# Backward-compatible with earlier versions.
read_eml = read
save_eml = save

__all__ = ['get_attachment', 'get_html', 'show', 'read', 'save', 'server', 'decode', 'read_eml', 'save_eml']


def server(user, password, smtp_host=None, smtp_port=None, pop_host=None, pop_port=None, smtp_ssl=None, pop_ssl=None,
           smtp_tls=None, pop_tls=None, config=None, timeout=60, auto_add_to=True, auto_add_from=True):
    """A wrapper to MailServer.

    SMTP:
        server.send_mail([recipient,], mail)

    POP3:
        server.get_mail(which)
        server.get_mails(subject, sender, after, before)
        server.get_latest()
        server.get_info()
        server.stat()

    Parse mail:
        server.show(mail)
        server.get_attachment(mail)
    """
    if config:
        config_dict = get_enterprise_server_config(config)
        if config_dict:
            return MailServer(user, password, **config_dict, timeout=timeout, auto_add_to=auto_add_to,
                              auto_add_from=auto_add_from)
        else:
            logger.warning('User-defined server config error, use default config instead.')

    return MailServer(**{k: v for k, v in locals().items() if k not in ('config_dict', 'config')})


def decode(mail_as_bytes, which=1):
    """Decode bytes mail, as usual in your disk."""
    return mail_decode(mail_as_bytes, which)



"""
zmail.exceptions
~~~~~~~~~~~~~~~~
Include all exceptions used within zmail.
"""


class ZmailException(IOError):
    """Base exception."""


class InvalidProtocol(ZmailException, ValueError):
    """Invalid protocol settings used."""
